# Copyright (c) 2008 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id$

from zope.interface import Interface, Attribute
from zope.publisher.interfaces.browser import IDefaultBrowserLayer

from silva.core import conf as silvaconf
from silva.core.interfaces import IMember

from Products.PluggableAuthService.interfaces.plugins import IExtractionPlugin


class IOpenIdExtractionPlugin(IExtractionPlugin):
    """Extract OpenID credential information from a request.
    """

    def initiateAuthentication(identity_url, return_to=None):
        """Initiate the OpenID authentication.
        """

    def extractCredentials(request):
        """ request -> { 'openid.identity' : identity,
                         'openid.assoc_handle' : assoc_handle,
                         'openid.return_to' : return_to,
                         'openid.signed' : signed,
                         'openid.sig' : sig,
                         'openid.invalidate_handle' : invalidate_handle,
                        }
        """


class IOpenIDAware(Interface):
    """Marker to let known that service_members is OpenID Aware.
    """


class IOpenIDAskUserInformation(Interface):
    """Let ask some information to the OpenID provider.
    """

    def require(name):
        """Require name from the provider.
        """

    def optional(name):
        """Ask optionally name to the provider.
        """


class IOpenIDAskedUserInformation(Interface):
    """Let OpenID plugin known which information have been asked.
    """

    def require():
        """Return a list of required information.
        """

    def optional():
        """Return a list of optional information.
        """


class IOpenIDMember(IMember):
    """This is an OpenID Member.
    """

    identity_url = Attribute("Identity URL of the user")
    fully_registered = Attribute(
        "Tell if the user complete the registration procedure")
    application_date = Attribute("Date when the user starts his registration")
    last_login_date = Attribute("Last login date of the user")


class ILoginPage(IDefaultBrowserLayer):
    """Resources for the OpendID login page
    """
    silvaconf.resource('openid.css')


class IRegisterPage(IDefaultBrowserLayer):
    """Resources for the OpendID registration page
    """
    silvaconf.resource('openid.css')



# Copyright (c) 2008 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id$

from five import grok
from zope.component import getUtility
from zope.component import getMultiAdapter, queryMultiAdapter
from zope.traversing.browser import absoluteURL

from silva.core.services.interfaces import IMemberService
from silva.core.views import views as silvaviews
from silva.core.views.interfaces import IHTTPResponseHeaders
from silva.pas.openid.interfaces import IOpenIDAskUserInformation
from silva.pas.openid.interfaces import ILoginPage, IRegisterPage
from silva.translations import translate as _

from zExceptions import BadRequest

from silva.fanstatic import need


class LoginPage(silvaviews.Page):
    grok.name('silva_login_form_with_openid.html')
    grok.require('zope2.Public')

    message = None
    action = None

    def update(self):
        need(ILoginPage)
        self.url = absoluteURL(self.context, self.request)
        if self.action is None:
            raise BadRequest()
        # Due how PAS monkey patch Zope, we need to do this by hand here.
        headers = queryMultiAdapter(
            (self.request, self),
            IHTTPResponseHeaders)
        if headers is not None:
            headers()


class RegisterPage(silvaviews.Page):
    grok.name('openid_register_new_user.html')
    grok.require('zope2.Public')

    def _process(self, captcha):
        self.feedback = None

        def get_form_value(key, default=None):
            return self.request.form.get(
                '__ac.field.identity.{0}'.format(key),
                default)

        if not get_form_value('do'):
            return
        identity = get_form_value('register', '').strip()
        if not identity:
            self.feedback = _(u"You need to provide an identity.")
            return
        verification = get_form_value('captcha', None)
        if verification is None:
            self.feedback = _(u"You need to fill the captcha.")
            return
        if not captcha.verify(verification):
            self.feedback = _(u"Error while validating the captcha.")
            return

        # Insert the URL under the name used by the PAS plugin in the
        # request to trigger the authentication when validating
        # credentials again.
        self.request.form['__ac.field.identity.url'] = identity

        root = self.context.get_root()
        service = getUtility(IMemberService)
        if not service.is_user(identity):
            # Create a member.
            members = root.Members
            factory = members.manage_addProduct['silva.pas.openid']
            factory.manage_addOpenIDMember(identity)
            # Ask information in order to fill-in member object.
            provider = IOpenIDAskUserInformation(self.request)
            provider.require('email')
            provider.optional('nickname')
            provider.optional('fullname')

        root.acl_users.validate(self.request)
        self.feedback = _(u'This URL is not a OpenID identity.')

    def update(self):
        need(IRegisterPage)
        captcha = getMultiAdapter((self.context, self.request), name='captcha')
        self._process(captcha)
        self.url = absoluteURL(self.context, self.request)
        self.captcha_img = captcha.image_tag()
        self.captcha_audio_url = captcha.audio_url()



# Copyright (c) 2008 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id$

from AccessControl.Permissions import manage_users as ManageUsers
from silva.core import conf as silvaconf

import install

silvaconf.extension_name("silva.pas.openid")
silvaconf.extension_title("Silva OpenID Support")
silvaconf.extension_depends(["Silva", 'silva.pas.base'])


def initialize(context):
    from .plugins import oid
    context.registerClass(oid.OpenIdPlugin,
                          permission=ManageUsers,
                          constructors=
                          (oid.manage_addOpenIDPluginForm,
                           oid.manage_addOpenIDPlugin),
                          visibility=None,
                          icon="www/openid.png")



# Copyright (c) 2008 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id$

from zope.interface import implements
from silva.pas.base.interfaces import IUserConverter

import urlparse

class OpenIDUserConverter(object):
    """ This let OpenID work, since getUserName and getUserId are
     called indifferently in Silva, but doesn't refere to the
     same data.
     """

    implements(IUserConverter)

    def match(self, userid):
        if not userid:
            return False
        ids = urlparse.urlparse(userid)
        if ids[1]:
            return True
        return False

    def convert(self, userid):
        if not userid:
            return userid
        ids = urlparse.urlparse(userid)
        real = ids[1]
        if len(ids[2]) > 1:
            real += '_' + str(ids[2][1:]).replace('/', '-')
        return real




# Copyright (c) 2008 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id$

import urlparse

# Zope 2
from AccessControl.SecurityInfo import ClassSecurityInfo
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from zExceptions import Redirect
import transaction
from zope.component import getUtility
from zope.component.event import objectEventNotify

# PAS
from Products.PluggableAuthService.plugins.BasePlugin import BasePlugin
from Products.PluggableAuthService.utils import classImplements
from Products.PluggableAuthService.interfaces.plugins import IAuthenticationPlugin
from Products.PluggableAuthService.PluggableAuthService import registerMultiPlugin


# Silva
from silva.pas.openid.interfaces import *
from silva.pas.base.interfaces import IUserConverter
from silva.pas.openid.store import ZopeStore
from silva.pas.openid.events import *
from silva.core.views.interfaces import IVirtualSite

# OpenID
from openid.yadis.discover import DiscoveryFailure
from openid.consumer.consumer import Consumer, SUCCESS

# Python
import logging

manage_addOpenIDPluginForm = PageTemplateFile("../www/openIDAddForm",
                globals(), __name__="manage_addOpenIDPluginForm")

logger = logging.getLogger("PluggableAuthService")

def manage_addOpenIDPlugin(self, id, title='', REQUEST=None):
    """Add a OpenID plugin to a Pluggable Authentication Service.
    """
    plugin = OpenIdPlugin(id, title)
    self._setObject(plugin.getId(), plugin)

    if REQUEST is not None:
        REQUEST["RESPONSE"].redirect("%s/manage_workspace"
                "?manage_tabs_message=OpenID+plugin+added." %
                self.absolute_url())


class OpenIdPlugin(BasePlugin):
    """OpenID authentication plugin.
    """

    meta_type = "OpenID plugin"
    security = ClassSecurityInfo()

    _properties = BasePlugin._properties + ({'id': 'policy_url',
                                             'type': 'string',
                                             'mode': 'w',
                                             'label': 'Policy URL',
                                             },)

    policy_url = ''

    def __init__(self, id, title=None):
        self._setId(id)
        self.title = title
        self.store = ZopeStore()

    def getConsumer(self):
        session=self.REQUEST["SESSION"]
        return Consumer(session, self.store)

    def extractOpenIdServerResponse(self, request, creds):
        """Process incoming redirect from an OpenId server.

        The redirect is detected by looking for the openid.mode
        form parameters. If it is found the creds parameter is
        cleared and filled with the found credentials.
        """

        mode = request.form.get("openid.mode", None)
        if mode == "id_res":
            # id_res means 'positive assertion' in OpenID, more commonly
            # describes as 'positive authentication'
            creds["openid.creds"] = request.form.copy()
        elif mode == "cancel":
            # cancel is a negative assertion in the OpenID protocol,
            # which means the user did not authorize correctly.
            objectEventNotify(OpenIDResultCancel(self, None, None))

    # IOpenIdExtractionPlugin implementation
    def initiateChallenge(self, request, identity_url, return_to=None):
        consumer = self.getConsumer()
        try:
            result = consumer.begin(identity_url)
        except DiscoveryFailure, e:
            logger.info("openid consumer discovery error for identity %s: %s",
                    identity_url, e[0])
            return
        except KeyError, e:
            logger.info("openid consumer error for identity %s: %s",
                    identity_url, e.why)
            pass

        site_url = IVirtualSite(request).get_top_level_url()
        site_parts = urlparse.urlparse(site_url)

        if return_to is None:
            return_to = request.form.get("__ac.field.origin", None)
            if return_to:
                # Complicated smash-up to build a full-url
                return_parts = urlparse.urlparse(return_to)
                result_parts = ()
                if not (return_parts[0] and return_parts[1]):
                    result_parts = result_parts + (
                        site_parts[0], site_parts[1])
                else:
                    result_parts = result_parts + (
                        return_parts[0], return_parts[1])
                result_parts = result_parts + (
                    return_parts[2], '', '', return_parts[5])
                return_to = urlparse.urlunparse(result_parts)
        if not return_to:
            return_to = site_url

        request["SESSION"]['return_to'] = return_to

        if self.policy_url:
            result.addExtensionArg('sreg', 'policy_url', self.policy_url)

        extra = IOpenIDAskedUserInformation(request)
        if extra.require():
            result.addExtensionArg('sreg',
                                   'required',
                                   ','.join(extra.require()))
        if extra.optional():
            result.addExtensionArg('sreg',
                                   'optional',
                                   ','.join(extra.optional()))

        url = result.redirectURL(site_url, return_to)

        # There is evilness here: we can not use a normal RESPONSE.redirect
        # since further processing of the request will happily overwrite
        # our redirect. So instead we raise a Redirect exception, However
        # raising an exception aborts all transactions, which means are
        # session changes are not stored. So we do a commit ourselves to
        # get things working.
        # XXX this also f**ks up ZopeTestCase
        transaction.commit()
        raise Redirect, url

    # IExtractionPlugin implementation
    def extractCredentials(self, request):
        """This method performs the PAS credential extraction.

        It takes either the zope cookie and extracts openid credentials
        from it, or a redirect from a OpenID server.
        """
        creds = {}
        identity = request.form.get("__ac.field.identity.url", '').strip()
        if identity:
            self.initiateChallenge(request, identity)
            return creds

        self.extractOpenIdServerResponse(request, creds)
        return creds


    # IAuthenticationPlugin implementation
    def authenticateCredentials(self, credentials):
        if not credentials.has_key("openid.creds"):
            return None
        real_creds = credentials["openid.creds"]
        if real_creds:
            consumer = self.getConsumer()
            result = consumer.complete(real_creds,
                                       self.REQUEST["SESSION"]['return_to'])
            userid = self._identityToId(result.identity_url)

            if result.status == SUCCESS:
                pas = self._getPAS()
                pas.updateCredentials(self.REQUEST,
                                      self.REQUEST.RESPONSE,
                                      userid,
                                      "")

                objectEventNotify(OpenIDResultSuccess(self, result, userid))
                return (userid, result.identity_url)
            else:
                logger.info("OpenId Authentication for %s failed: %s",
                            result.identity_url, result.message)
                objectEventNotify(OpenIDResultCancel(self, result, userid))

        return None

    def _identityToId(self, identity):
        utility = getUtility(IUserConverter, name="openid")
        converter = utility()
        return converter.convert(identity)


classImplements(OpenIdPlugin, IOpenIdExtractionPlugin, IAuthenticationPlugin)
registerMultiPlugin(OpenIdPlugin.meta_type)



# Copyright (c) 2008 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id$

from zope.interface import Interface, Attribute, implements


class IOpenIDResult(Interface):
    """Event triggered when a OpenID Result occurs.
    """

    object = Attribute("Originate plugin")
    result = Attribute("Result of the OpenID request")

class IOpenIDResultSuccess(IOpenIDResult):
    """The OpenID result was a success.
    """

    userid = Attribute("UserID")


class OpenIDResult(object):
    implements(IOpenIDResult)

    def __init__(self, plugin, result):
        self.object = plugin
        self.result = result


class OpenIDResultSuccess(OpenIDResult):

    implements(IOpenIDResultSuccess)

    def __init__(self, plugin, result, userid):
        super(OpenIDResultSuccess, self).__init__(plugin, result)
        self.userid = userid


class IOpenIDResultCancel(IOpenIDResult):
    """The OpenID result was a cancel
    """

class OpenIDResultCancel(OpenIDResult):

    implements(IOpenIDResultCancel)



# Copyright (c) 2008 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id$

from interfaces import IOpenIDMember
import DateTime

def atOpenIDSuccess(object, event):
    root = object.get_root()
    member = getattr(root.Members, event.userid)

    assert IOpenIDMember.providedBy(member)

    if not member.fully_registered:
        extra = event.result.extensionResponse('sreg', True)
        if extra.has_key('email'):
            member.set_email(extra['email'])
        if extra.has_key('fullname'):
            member.set_fullname(extra['fullname'])
        if extra.has_key('title'):
            member._title = extra['title']
        member.fully_registered = True
    member.last_login_date = DateTime.DateTime()

def atOpenIDCancel(object, event):
    if event.userid:
        root = object.get_root()
        member = getattr(root.Members, event.userid)

        assert IOpenIDMember.providedBy(member)

        if not member.fully_registered:
            root.Members.manage_delObjects([event.userid,])




# Copyright (c) 2008 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id$

from five import grok
from zope.component import getUtility

from AccessControl import ClassSecurityInfo
from App.class_init import InitializeClass
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
import DateTime

from Products.Silva.SimpleMembership import SimpleMember
from Products.Silva.helpers import add_and_edit
from Products.Silva import SilvaPermissions

from silva.core import conf as silvaconf
from silva.pas.base.interfaces import IUserConverter
from interfaces import IOpenIDMember


class SilvaOpenIDMember(SimpleMember):
    grok.implements(IOpenIDMember)

    meta_type = 'Silva OpenID Member'
    security = ClassSecurityInfo()
    silvaconf.icon('www/member.png')
    silvaconf.factory('manage_addOpenIDMemberForm')
    silvaconf.factory('manage_addOpenIDMember')

    manage_options = ({'label': 'Details',
                       'action': 'manage_details',
                       },) + SimpleMember.manage_options

    def __init__(self, identity_url, id):
        super(SilvaOpenIDMember, self).__init__(id)
        self.identity_url = identity_url
        self.application_date = DateTime.DateTime()
        self.fully_registered = False
        self.last_login_date = None

    security.declareProtected(SilvaPermissions.ViewManagementScreens,
                              'manage_details')
    manage_details = PageTemplateFile(
        "www/openIDMemberDetails", globals(),
        __name__='manage_details')

    manage_main = manage_details


InitializeClass(SilvaOpenIDMember)


manage_addOpenIDMemberForm = PageTemplateFile(
    "www/openIDMemberAdd", globals(),
    __name__='manage_addOpenIDMemberForm')


def manage_addOpenIDMember(self, identity_url, REQUEST=None):
    """Add a OpenID Member.
    """
    utility = getUtility(IUserConverter, name="openid")
    converter = utility()
    userid = converter.convert(identity_url)

    user = SilvaOpenIDMember(identity_url, userid)
    self._setObject(userid, user)
    user = getattr(self, userid)
    user.manage_addLocalRoles(id, ['ChiefEditor'])
    add_and_edit(self, userid, REQUEST)
    return ''



# Copyright (c) 2008 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id$

from setuptools import setup, find_packages
import os

version = '3.0'

setup(name='silva.pas.openid',
      version=version,
      description="OpenID suport for Silva CMS",
      long_description= open('README.txt').read() + "\n" +
                       open(os.path.join("docs", "HISTORY.txt")).read(),
      classifiers=[
        "Framework :: Zope2",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: BSD License",
        ],
      keywords='pas openid silva',
      author='Sylvain Viollon',
      author_email='info@infrae.com',
      url='https://github.com/silvacms/silva.pas.openid',
      license='BSD',
      package_dir={'': 'src'},
      packages=find_packages('src', exclude=['ez_setup']),
      namespace_packages=['silva', 'silva.pas'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
        "python-openid >= 2.1",
        "five.grok",
        "silva.core.interfaces",
        "silva.core.conf",
        "silva.core.views",
        "silva.core.services",
        "silva.pas.base",
        "silva.pas.membership",
        "silva.captcha",
        "setuptools",
        "zope.component",
        "zope.interface",
        "zope.publisher"],
      )



# Copyright (c) 2008 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id$

from Products.Silva.tests import SilvaTestCase
from Products.Five import zcml

from Testing import ZopeTestCase as ztc


class OpenIDDependencyTestCase(SilvaTestCase.SilvaTestCase):
    """Test that if you try to use OpenID without the dependency
    installed in service extensions, you get an error.
    """

    def test_install(self):

        def install():
            root = self.getRoot()
            root.service_extensions.install('silva.pas.openid')

        self.assertRaises(AssertionError, install)


class OpenIDTestCase(SilvaTestCase.SilvaTestCase):
    """Test case for OpenID implementation.
    """

    def afterSetUp(self):
        """After set up, install the extension.
        """
        root = self.getRoot()
        root.service_extensions.install('silva.pas.base')
        root.service_extensions.install('silva.pas.openid')


    def test_00install(self):
        """Test install.
        """
        root = self.getRoot()

        # First the extension should be installed
        service_extensions = root.service_extensions
        self.failUnless(service_extensions.is_installed('silva.pas.base'))
        self.failUnless(service_extensions.is_installed('silva.pas.openid'))


    def test_99uninstall(self):
        """Test uninstall.
        """

        root = self.getRoot()
        root.service_extensions.uninstall('silva.pas.openid')

        # Should be uninstalled now
        service_extensions = root.service_extensions
        self.failIf(service_extensions.is_installed('silva.pas.openid'))

        # But we can reinstall it
        root.service_extensions.install('silva.pas.openid')
        self.failUnless(service_extensions.is_installed('silva.pas.openid'))


import unittest
def test_suite():

    # Load Five ZCML
    from Products import Five
    zcml.load_config('meta.zcml', Five)
    zcml.load_config('configure.zcml', Five)

    # Load our ZCML, which add the extension as a Product
    from silva.pas import openid
    zcml.load_config('configure.zcml', openid)

    # Load the Zope Product
    ztc.installProduct('GenericSetup')
    ztc.installProduct('PluginRegistry')
    ztc.installProduct('PluggableAuthService')
    ztc.installPackage('plone.session')
    ztc.installPackage('silva.pas.base')
    ztc.installPackage('silva.pas.membership')
    ztc.installPackage('silva.pas.openid')

    # Run tests
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(OpenIDDependencyTestCase))
    suite.addTest(unittest.makeSuite(OpenIDTestCase))
    return suite



# Copyright (c) 2008 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id$

from zope.publisher.interfaces.browser import IBrowserRequest
from zope.component import adapts
from zope.interface import implements

from interfaces import IOpenIDAskUserInformation, IOpenIDAskedUserInformation


REQUIRE_KEY = 'openid-require'
OPTIONAL_KEY = 'openid-optional'

class AnnotatedAdapter(object):

    def __init__(self, request):
        self.context = request
        # following is ugly but request is not annotable
        if not hasattr(request, '__annotations__'):
            setattr(request, '__annotations__', dict())
        self.annotations = request.__annotations__

class AskOpenIDOnRequest(AnnotatedAdapter):

    implements(IOpenIDAskUserInformation)
    adapts(IBrowserRequest)

    def require(self, name):
        requires = self.annotations.get(REQUIRE_KEY, [])
        requires.append(name)
        self.annotations[REQUIRE_KEY] = requires

    def optional(self, name):
        optionals = self.annotations.get(OPTIONAL_KEY, [])
        optionals.append(name)
        self.annotations[OPTIONAL_KEY] = optionals


class AskedOpenIDOnRequest(AnnotatedAdapter):

    implements(IOpenIDAskedUserInformation)
    adapts(IBrowserRequest)

    def require(self):
        return self.annotations.get(REQUIRE_KEY, [])

    def optional(self):
        return self.annotations.get(OPTIONAL_KEY, [])





# Copyright (c) 2008 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id$

from zope.interface import alsoProvides, noLongerProvides

from Products.PluggableAuthService.interfaces import plugins
from silva.pas.base.interfaces import IPASService
from silva.pas.openid.interfaces import IOpenIDAware


def install(root, extension):
    """Installation method for OpenID support
    """
    assert IPASService.providedBy(root.service_members), \
        u"This extension requires silva.pas.base"

    # Register PAS plugins
    registerPASPlugins(root.acl_users)

    alsoProvides(root.service_members, IOpenIDAware)


def uninstall(root, extension):
    """Uninstall OpenID support
    """
    assert IPASService.providedBy(root.service_members), \
        u"This extension requires silva.pas.base"
    unregisterPASPlugins(root.acl_users)
    noLongerProvides(root.service_members, IOpenIDAware)


def is_installed(root, extension):
    return IOpenIDAware.providedBy(root.service_members)


def registerPASPlugins(pas):
    """Register new PAS plugins.
    """
    plugin_ids = pas.objectIds()
    assert 'cookie_auth' in plugin_ids, \
        'Acl user have not been created by silva.pas.base'
    if 'openid' not in plugin_ids:
        factory = pas.manage_addProduct['silva.pas.openid']
        factory.manage_addOpenIDPlugin('openid')
    if 'members' not in plugin_ids:
        factory = pas.manage_addProduct['silva.pas.membership']
        factory.manage_addMembershipPlugin('members')

    def registerPluginIfNew(ptype, pid):
        if pid not in pas.plugins.listPluginIds(ptype):
            pas.plugins.activatePlugin(ptype, pid)

    registerPluginIfNew(plugins.IExtractionPlugin, 'openid')
    registerPluginIfNew(plugins.IAuthenticationPlugin, 'openid')
    registerPluginIfNew(plugins.IUserEnumerationPlugin, 'members')
    registerPluginIfNew(plugins.ICredentialsUpdatePlugin, 'cookie_auth')
    # Change login form
    pas.cookie_auth.login_path = 'silva_login_form_with_openid.html'


def unregisterPASPlugins(pas):
    """Remove PAS plugins.
    """
    plugin_ids = pas.objectIds()
    assert 'cookie_auth' in plugin_ids, \
        'Acl user have not been created by silva.pas.base'
    if 'openid' not in plugin_ids:
        pas.manage_delObjects(['openid'])
    if 'members' not in plugin_ids:
        pas.manage_delObjects(['members'])

    pas.manage_delObjects(['openid', 'members',])
    # Restore login form to default one
    pas.cookie_auth.login_path = 'silva_login_form.html'


if __name__ == '__main__':
    print """This module is not an installer. You don't have to run it."""
